'' =================================================================================================
''
''   File....... jm_sprite.spin
''   Purpose.... Serial Interface for MedeaWiz Sprite Video player
''               -- based on FullDuplexSerial by Chip Gracey
''   Author..... Jon "JonnyMac" McPhalen
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 15 MAR 2017
''
'' =================================================================================================


con { io pins }

  RX1  = 31                                                     ' programming / terminal
  TX1  = 30

  SDA1 = 29                                                     ' eeprom / i2c
  SCL1 = 28


con { sprite commands / responses }

  KEY_MUTE   = $E8
  KEY_VOL_UP = $E9
  KEY_VOL_DN = $EA
  
  KEY_PAUSE  = $F0
  KEY_REW    = $F1
  KEY_FWD    = $F2
  KEY_PREV   = $F3
  KEY_NEXT   = $F4

  EOF        = $EE


var

  long  cog

  long  buffer
  long  ms


pub null

  ' This is not a top-level object


pub start(rxp, txp, baudrate)

'' Start UART for Sprite video player

  stop

  buffer.byte[0] := rxp                                         ' set pins
  buffer.byte[1] := txp
  buffer.word[1] := clkfreq / baudrate                          ' ticks per bit

  ms := clkfreq / 1_000                                         ' ticks in 1ms

  cog := cognew(@entry, @buffer) + 1                            ' launc the sprite uart

  return cog


pub stop

  if (cog)
    cogstop(cog - 1)
    cog := 0
    
  buffer := EOF << 16                                           ' spoof EOF rx
  ms := 0                                                       ' reset run time to 0

  
pub busy

'' Returns true if file is playing

  if (buffer.byte[2] =< 200)
    return true
  else
    return false


pub current_file

'' Returns current file (0..200) or EOF ($EE) 

  return buffer.byte[2]


pub alert

'' Returns special status code

  return buffer.byte[3]


pub command(c)

'' Send command byte to UART

  buffer.word[0] := HAS_CMD | (c & $FF)


pub millis

  return ms #> 0                                                ' 0 if not playing
  

dat  { unbuffered fd uart with timer }

                        org     0

entry
rxcode  if_never        mov     rxcode, #receive                ' assign task addresses 
txcode  if_never        mov     txcode, #transmit                       
mscode  if_never        mov     mscode, #ms_timer

                        rdlong  t1, par                         ' get pins & bit timing
                        mov     t2, t1                          ' make a copy

                        and     t1, #$1F                        ' setup rx
                        mov     rxmask, #1
                        shl     rxmask, t1

                        mov     t1, t2                          ' setup tx
                        shr     t1, #8
                        and     t1, #$1F   
                        mov     txmask, #1 
                        shl     txmask, t1
                        or      outa, txmask                    ' set tx to idle
                        or      dira, txmask

                        mov     bittix, t2                      ' setup bittix
                        shr     bittix, #16

                        mov     rxhub, par                      ' setup hub addresses
                        add     rxhub, #2                       ' rx to buffer.byte[2]
                        mov     mshub, par
                        add     mshub, #4

                        rdlong  mstix, mshub                    ' get ticks/ms

                        mov     t1, #EOF                        ' initialize buffer & timer
                        shl     t1, #16
                        wrlong  t1, par
                        
                        mov     msruntime, #0                   ' clear hub timer
                        wrlong  msruntime, mshub
          
                        mov     flags, #%000                    ' not playing now
                                

' -----------------                                              
'  Receive Process                                               
' ----------------- 

receive                 jmpret  rxcode, txcode
                                   
                        test    rxmask, ina             wc      ' look for start bit      
        if_c            jmp     #receive                         
                                                                 
                        mov     rxbits, #9                      ' ready to receive byte
                        mov     rxcnt, bittix                    
                        shr     rxcnt, #1                        
                        add     rxcnt, cnt                           
                                                                  
:bit                    add     rxcnt, bittix                   ' ready next bit period
                                                                  
:wait                   jmpret  rxcode, txcode
                                                                  
                        mov     t1, rxcnt                       ' check if bit receive period done
                        sub     t1, cnt                           
                        cmps    t1, #0                  wc        
        if_nc           jmp     #:wait                            
                                                                  
                        test    rxmask, ina             wc      ' receive bit on rx pin
                        rcr     rxdata, #1                       
                        djnz    rxbits, #:bit                    
                                                                 
                        shr     rxdata, #32-9                   ' justify and trim received byte
                        and     rxdata, #$FF                    ' clean-up
                        cmp     rxdata, #200            wz, wc  ' check for valid file
        if_a            jmp     #:end_play                      ' check for EOF/special status

                        wrword  rxdata, rxhub                   ' write file, clear special
                        jmpret  rxcode, txcode
                        
                        test    flags, #%100            wc      ' save pause flag
                        shl     flags, #1                       ' make room for update
                        or      flags, #1                       ' still playing
                        and     flags, #%11                     ' clean-up
                        cmp     flags, #%01             wz      ' new start?
                        muxc    flags, #%100                    ' restore pause flag   
        if_ne           jmp     #receive
       
                        mov     mscnt, cnt
                        add     mscnt, mstix
                        mov     msruntime, #0  
                        wrlong  msruntime, mshub                                               
                        jmp     #receive

:end_play               cmp     rxdata, #EOF            wz      ' end-of-file?
        if_e            mov     flags, #%000                    ' reset flags on EOF
        if_ne           shl     rxdata, #8                      ' other byte, shift
                        wrword  rxdata, rxhub 
                        jmp     #receive   
 

' ------------------                                              
'  Transmit Process                                               
' ------------------ 

transmit                jmpret  txcode, mscode

                        rdword  txdata, par                     ' check for command
                        test    txdata, HAS_CMD         wc
        if_nc           jmp     #transmit

                        wrword  ZERO, par

                        mov     txlast, txdata                  ' save clean copy
                        and     txlast, #$FF      

                        or      txdata, STOP_BITS               ' add stop bit(s)
                        shl     txdata, #1                      ' add start bit
                        mov     txbits, #11                     ' bits = start + 8 + 2 stop
                        mov     txcnt, cnt                      ' start bit timer 
                      
:bit                    shr     txdata, #1              wc      ' get bit from txdata    
                        muxc    outa, txmask                    ' move to tx pin                   
                        add     txcnt, bittix                   ' ready next cnt
                                                                 
:wait                   jmpret  txcode, mscode
                                                                 
                        mov     t1, txcnt                       ' check if bit transmit period done
                        sub     t1, cnt                          
                        cmps    t1, #0                  wc       
        if_nc           jmp     #:wait                           
                                                                 
                        djnz    txbits, #:bit                   ' another bit to transmit?
                            
                        cmp     txlast, #200            wc, wz  ' file command?
        if_a            jmp     #:chk_pause

                        mov     mscnt, cnt                      ' restart timing on new file cmd          
                        add     mscnt, mstix         
                        mov     msruntime, #0        
                        wrlong  msruntime, mshub     
                        jmp     #transmit
                                 
:chk_pause              cmp     txlast, #KEY_PAUSE      wz
        if_e            xor     flags, #%100                    ' toggle pause flag
                        jmp     #transmit 
 

' -------------------
'  1ms Timer Process
' -------------------
                        
ms_timer                jmpret  mscode, rxcode
                        mov     t1, mscnt
                        sub     t1, cnt
                        cmps    t1, #0                  wc
        if_nc           jmp     #ms_timer
        
                        add     mscnt, mstix                    ' set new target 
                        test    flags, #%100            wc      ' set c if paused
                        test    flags, #%011            wz      ' set z if not playing
        if_nc_and_nz    add     msruntime, #1                   ' running -- inc ms
        if_nc_and_nz    wrlong  msruntime, mshub
                        jmp     #ms_timer


' --------------------------------------------------------------------------------------------------

ZERO                    long    0
HAS_CMD                 long    $1_00
STOP_BITS               long    $FFFF_FF00

bittix                  res     1                               ' ticks per bit
                                                                 
rxmask                  res     1                               ' mask for rx pin
rxhub                   res     1                               ' hub address of rxbuffer                             ' 
rxdata                  res     1                               ' received byte
rxbits                  res     1                               ' bit counter for rx
rxcnt                   res     1                               ' bit timer for rx
                                                                 
txmask                  res     1                               ' mask for tx pin
txdata                  res     1                               ' byte to transmit
txlast                  res     1                               ' last byte transmitted
txbits                  res     1                               ' bit counter for tx
txcnt                   res     1                               ' bit timer for tx

mshub                   res     1                               ' hub address of ms
mstix                   res     1                               ' ticks/millisecond
msruntime               res     1                               ' run time of current file in ms
mscnt                   res     1                               ' timer for 1ms

flags                   res     1
                                                                 
t1                      res     1                               ' working registers
t2                      res     1                                
                        
                        fit     496

    
dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      